/*
 * Altogether: Xerox Alto microcode-level simulator
 * Diablo 31, 44 disk and controller simulation
 * $Id: disk.c 93 2003-10-24 02:04:52Z eric $
 * Copyright 2001 Eric Smith <eric@brouhaha.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  Note that permission is
 * not granted to redistribute this program under the terms of any
 * other version of the General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111  USA
 */

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include "alto.h"
#include "debug.h"
#include "cpu.h"
#include "scheduler.h"
#include "disk.h"


typedef struct
{
  char *name;
  int sectors_per_track;
  double rpm;
  double bit_rate;
} drive_type_info_t;


static drive_type_info_t drive_type_info [MAX_DRIVE_TYPE] =
{
  { "Diablo 31", 12, 1500, (3.330e6 / 2) },
  { "Diablo 44", 12, 2400, (5.000e6 / 2) },
};


static drive_type_t drive_type;
static sim_time_t word_time, sector_time;


static int selected_drive;
static bool disk_in_place [MAX_DRIVE];
static int current_sector [MAX_DRIVE];


static uint16_t kstat;
static uint16_t kcmd;


#define bs_disk_kstat       bs_task_3
#define bs_disk_kdata       bs_task_4

#define f1_disk_strobe      f1_task_11
#define f1_disk_load_kstat  f1_task_12
#define f1_disk_increcno    f1_task_13
#define f1_disk_clrstat     f1_task_14
#define f1_disk_load_kcomm  f1_task_15
#define f1_disk_load_kadr   f1_task_16
#define f1_disk_load_kdata  f1_task_17

#define f2_disk_init        f2_task_10
#define f2_disk_rwc         f2_task_11
#define f2_disk_recno       f2_task_12
#define f2_disk_xfrdat      f2_task_13
#define f2_disk_swrnrdy     f2_task_14
#define f2_disk_nfer        f2_task_15
#define f2_disk_strobon     f2_task_16


static void sector_mark (uint32_t arg1, void *arg2)
{
  if (! disk_in_place [arg1])
    return;
  current_sector [arg1]++;
  if (current_sector [arg1] >= drive_type_info [drive_type].sectors_per_track)
    current_sector [arg1] = 0;
  if (arg1 == selected_drive)
    set_task_request (task_disk_sector);
  schedule_event (current_time + sector_time, sector_mark, arg1, NULL);
}


static void select_drive (int drive)
{
  selected_drive = drive;
}


static void insert_disk (int drive)
{
  if (! disk_in_place [drive])
    schedule_event (sector_time, sector_mark, drive, NULL);
  disk_in_place [drive] = true;
}


static void remove_disk (int drive)
{
#if 0
  if (disk_in_place [drive])
    unschedule_event (?);
#endif
  disk_in_place [drive] = false;
}


void bs_fn_disk_kstat (int task, int code, int after)
{
  if (! after)
    drive_bus ("kstat", kstat);  /* $$$ merge in current sector */
}

void bs_fn_disk_kdata (int task, int code, int after)
{
  /* ??? */
}

void f1_fn_disk_strobe (int task, int code, int after)
{
  /* ??? */
}

void f1_fn_disk_load_kstat (int task, int code, int after)
{
  /* ??? */
}

void f1_fn_disk_increcno (int task, int code, int after)
{
  /* ??? */
}

void f1_fn_disk_clrstat (int task, int code, int after)
{
  /* ??? */
}

void f1_fn_disk_load_kcomm (int task, int code, int after)
{
  /* ??? */
}

void f1_fn_disk_load_kadr (int task, int code, int after)
{
  /* ??? */
}

void f1_fn_disk_load_kdata (int task, int code, int after)
{
  /* ??? */
}

void f2_fn_disk_init (int task, int code, int after)
{
  /* ??? */
}

void f2_fn_disk_rwc (int task, int code, int after)
{
  /* ??? */
}

void f2_fn_disk_recno (int task, int code, int after)
{
  /* ??? */
}

void f2_fn_disk_xfrdat (int task, int code, int after)
{
  /* ??? */
}

void f2_fn_disk_swrnrdy (int task, int code, int after)
{
  /* ??? */
}

void f2_fn_disk_nfer (int task, int code, int after)
{
  /* ??? */
}

void f2_fn_disk_strobon (int task, int code, int after)
{
  /* ??? */
}


void install_disk (void)
{
  install_bus_select_fn (task_disk_word, bs_disk_kstat, & bs_fn_disk_kstat);
  install_bus_select_fn (task_disk_word, bs_disk_kdata, & bs_fn_disk_kdata);

  install_f1_fn (task_disk_word, f1_disk_strobe,     & f1_fn_disk_strobe);
  install_f1_fn (task_disk_word, f1_disk_load_kstat, & f1_fn_disk_load_kstat);
  install_f1_fn (task_disk_word, f1_disk_increcno,   & f1_fn_disk_increcno);
  install_f1_fn (task_disk_word, f1_disk_clrstat,    & f1_fn_disk_clrstat);
  install_f1_fn (task_disk_word, f1_disk_load_kcomm, & f1_fn_disk_load_kcomm);
  install_f1_fn (task_disk_word, f1_disk_load_kadr,  & f1_fn_disk_load_kadr);
  install_f1_fn (task_disk_word, f1_disk_load_kdata, & f1_fn_disk_load_kdata);

  install_f2_fn (task_disk_word, f2_disk_init,       & f2_fn_disk_init);
  install_f2_fn (task_disk_word, f2_disk_rwc,        & f2_fn_disk_rwc);
  install_f2_fn (task_disk_word, f2_disk_recno,      & f2_fn_disk_recno);
  install_f2_fn (task_disk_word, f2_disk_xfrdat,     & f2_fn_disk_xfrdat);
  install_f2_fn (task_disk_word, f2_disk_swrnrdy,    & f2_fn_disk_swrnrdy);
  install_f2_fn (task_disk_word, f2_disk_nfer,       & f2_fn_disk_nfer);
  install_f2_fn (task_disk_word, f2_disk_strobon,    & f2_fn_disk_strobon);

  drive_type = DIABLO_31;

  word_time = 1e9 * (16 / drive_type_info [drive_type].bit_rate);
  sector_time = 1e9 * ((60.0 / drive_type_info [drive_type].rpm) /
		       drive_type_info [drive_type].sectors_per_track);

#if 0
  insert_disk (0);
#endif
}
