/*
 * Altogether: Xerox Alto microcode-level simulator
 * Priority Queue
 * $Id: priority_queue.c 93 2003-10-24 02:04:52Z eric $
 * Copyright 2001 Eric Smith <eric@brouhaha.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  Note that permission is
 * not granted to redistribute this program under the terms of any
 * other version of the General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111  USA
 */

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

#include "priority_queue.h"


typedef struct 
{
  priority_t priority;
  void *data;
} element_t;


struct priority_queue_t
{
  int max;
  int n;
  element_t *element;
};


priority_queue_handle new_priority_queue (int max_entries)
{
  priority_queue_handle qh;

  qh = calloc (1, sizeof (struct priority_queue_t));
  if (! qh)
    goto fail;

  qh->element = calloc (max_entries, sizeof (element_t));
  if (! qh->element)
    goto fail;

  qh->max = max_entries;
  return (qh);

 fail:
  if (qh)
    {
      if (qh->element)
	free (qh->element);
      free (qh);
    }
  return (NULL);
}

bool priority_queue_empty (priority_queue_handle qh)
{
  return (qh->n == 0);
}

bool priority_queue_full (priority_queue_handle qh)
{
  return (qh->n == qh->max);
}

int priority_queue_size (priority_queue_handle qh)
{
  return (qh->n);
}


#define left_child_of(n) (2 * (n) + 1)
#define right_child_of(n) (2 * (n) + 1)
#define parent_of(n) (((n) - 1) / 2)


static void swap (element_t *e1, element_t *e2)
{
  element_t temp;
  temp = *e1;
  *e1 = *e2;
  *e2 = temp;
}


bool priority_queue_add (priority_queue_handle qh,
			 priority_t priority, void *data)
{
  int i;

  if (qh->n == qh->max)
    return (false);
  qh->element [qh->n].priority = priority;
  qh->element [qh->n].data = data;

  i = qh->n;
  while ((i != 0) &&
	 (qh->element [i].priority < qh->element [parent_of (i)].priority))
    {
      swap (& qh->element [i], & qh->element [parent_of (i)]);
      i = parent_of (i);
    }
  qh->n++;
  return (true);
}


static void balance_down (priority_queue_handle qh, int i)
{
  int l;
  int r;

  for (;;)
    {
      l = left_child_of (i);
      if (l >= qh->n)
	return;

      r = right_child_of (i);

      if ((qh->element [l].priority < qh->element [i].priority) &&
	  ((r >= qh->n) || (qh->element [l].priority < qh->element [r].priority)))
	{
	  swap (& qh->element [i], & qh->element [l]);
	  i = l;
	  continue;
	}

      if ((r >= qh->n) || (qh->element [i].priority < qh->element [r].priority))
	return;

      swap (& qh->element [i], & qh->element [r]);
      i = r;
    }
}


bool priority_queue_peek (priority_queue_handle qh,
			  priority_t *priority, void **data)
{
  element_t result;

  if (qh->n == 0)
    return (false);

  result = qh->element [0];

  (* priority) = result.priority;
  (* data) = result.data;
  return (true);
}


bool priority_queue_remove (priority_queue_handle qh,
			    priority_t *priority, void **data)
{
  element_t result;

  if (qh->n == 0)
    return (false);

  result = qh->element [0];
  qh->n--;
  if (qh->n)
    {
      qh->element [0] = qh->element [qh->n];
      balance_down (qh, 0);
    }

  (* priority) = result.priority;
  (* data) = result.data;
  return (true);
}


priority_t priority_queue_head (priority_queue_handle qh)
{
  if (qh->n == 0)
    return (PQ_LOWEST_PRIORITY);
  return (qh->element [0].priority);
}

